"""Tests for ways the helper can be used."""

from http import HTTPStatus
from typing import Final

import httpretty  # pyright: ignore[reportMissingTypeStubs]
import httpx
import requests
import requests_mock as req_mock
import responses
import respx
from flask import Flask

from requests_mock_flask import add_flask_app_to_mock

# We use a high timeout to allow interactive debugging while requests are being
# made.
_TIMEOUT_SECONDS: Final[int] = 120


class TestResponses:
    """Tests for using the helper with ``responses``."""

    @staticmethod
    def test_context_manager() -> None:
        """
        It is possible to use the helper with a ``responses`` context
        manager.
        """
        app = Flask(import_name=__name__, static_folder=None)

        @app.route(rule="/")
        def _() -> str:
            """Return a simple message."""
            return "Hello, World!"

        response = requests.Response()

        with responses.RequestsMock(
            assert_all_requests_are_fired=False,
        ) as resp_m:
            add_flask_app_to_mock(
                mock_obj=resp_m,
                flask_app=app,
                base_url="http://www.example.com",
            )

            response = requests.get(
                url="http://www.example.com",
                timeout=_TIMEOUT_SECONDS,
            )

        assert response.status_code == HTTPStatus.OK
        assert response.text == "Hello, World!"

    @staticmethod
    @responses.activate
    def test_decorator() -> None:
        """It is possible to use the helper with a ``responses`` decorator."""
        app = Flask(import_name=__name__, static_folder=None)

        @app.route(rule="/")
        def _() -> str:
            """Return a simple message."""
            return "Hello, World!"

        add_flask_app_to_mock(
            mock_obj=responses,
            flask_app=app,
            base_url="http://www.example.com",
        )

        response = requests.get(
            url="http://www.example.com",
            timeout=_TIMEOUT_SECONDS,
        )

        assert response.status_code == HTTPStatus.OK
        assert response.text == "Hello, World!"


class TestRequestsMock:
    """Tests for using the helper with ``requests_mock``."""

    @staticmethod
    def test_context_manager() -> None:
        """
        It is possible to use the helper with a ``requests_mock``
        context
        manager.
        """
        app = Flask(import_name=__name__, static_folder=None)

        @app.route(rule="/")
        def _() -> str:
            """Return a simple message."""
            return "Hello, World!"

        with req_mock.Mocker() as resp_m:
            add_flask_app_to_mock(
                mock_obj=resp_m,
                flask_app=app,
                base_url="http://www.example.com",
            )

            response = requests.get(
                url="http://www.example.com",
                timeout=_TIMEOUT_SECONDS,
            )

        assert response.status_code == HTTPStatus.OK
        assert response.text == "Hello, World!"

    @staticmethod
    def test_fixture(requests_mock: req_mock.Mocker) -> None:
        """
        It is possible to use the helper with a ``requests_mock``
        fixture.
        """
        app = Flask(import_name=__name__, static_folder=None)

        @app.route(rule="/")
        def _() -> str:
            """Return a simple message."""
            return "Hello, World!"

        add_flask_app_to_mock(
            mock_obj=requests_mock,
            flask_app=app,
            base_url="http://www.example.com",
        )

        response = requests.get(
            url="http://www.example.com",
            timeout=_TIMEOUT_SECONDS,
        )

        assert response.status_code == HTTPStatus.OK

    @staticmethod
    def test_adapter() -> None:
        """
        It is possible to use the helper with a ``requests_mock``
        adapter.
        """
        app = Flask(import_name=__name__, static_folder=None)

        @app.route(rule="/")
        def _() -> str:
            """Return a simple message."""
            return "Hello, World!"

        session = requests.Session()
        adapter = req_mock.Adapter()
        session.mount(prefix="mock", adapter=adapter)

        add_flask_app_to_mock(
            mock_obj=adapter,
            flask_app=app,
            base_url="mock://www.example.com",
        )

        response = session.get(url="mock://www.example.com")

        assert response.status_code == HTTPStatus.OK
        assert response.text == "Hello, World!"


class TestHTTPretty:
    """Tests for using the helper with HTTPretty.

    We only test one way of using the helper with HTTPretty, because the
    other ways also pass in the HTTPretty module.
    """

    @staticmethod
    def test_use() -> None:
        """It is possible to use the helper with HTTPretty."""
        app = Flask(import_name=__name__, static_folder=None)

        @app.route(rule="/")
        def _() -> str:
            """Return a simple message."""
            return "Hello, World!"

        with httpretty.enabled():  # type: ignore[no-untyped-call]
            add_flask_app_to_mock(
                mock_obj=httpretty,
                flask_app=app,
                base_url="http://www.example.com",
            )

            response = requests.get(
                url="http://www.example.com",
                timeout=_TIMEOUT_SECONDS,
            )

        assert response.status_code == HTTPStatus.OK
        assert response.text == "Hello, World!"


class TestRespx:
    """Tests for using the helper with ``respx``."""

    @staticmethod
    def test_context_manager() -> None:
        """
        It is possible to use the helper with a ``respx`` context
        manager.
        """
        app = Flask(import_name=__name__, static_folder=None)

        @app.route(rule="/")
        def _() -> str:
            """Return a simple message."""
            return "Hello, World!"

        with respx.mock(assert_all_called=False) as respx_mock:
            add_flask_app_to_mock(
                mock_obj=respx_mock,
                flask_app=app,
                base_url="http://www.example.com",
            )

            response = httpx.get(url="http://www.example.com")

        assert response.status_code == HTTPStatus.OK
        assert response.text == "Hello, World!"

    @staticmethod
    @respx.mock(assert_all_called=False)
    def test_decorator(respx_mock: respx.MockRouter) -> None:
        """
        It is possible to use the helper with a ``respx``
        decorator.
        """
        app = Flask(import_name=__name__, static_folder=None)

        @app.route(rule="/")
        def _() -> str:
            """Return a simple message."""
            return "Hello, World!"

        add_flask_app_to_mock(
            mock_obj=respx_mock,
            flask_app=app,
            base_url="http://www.example.com",
        )

        response = httpx.get(url="http://www.example.com")

        assert response.status_code == HTTPStatus.OK
        assert response.text == "Hello, World!"



"""Setup for Sybil."""

from doctest import ELLIPSIS

import pytest
from beartype import beartype
from sybil import Sybil
from sybil.parsers.rest import DocTestParser, PythonCodeBlockParser


def pytest_collection_modifyitems(items: list[pytest.Item]) -> None:
    """Apply the beartype decorator to all collected test functions."""
    for item in items:
        if isinstance(item, pytest.Function):
            item.obj = beartype(obj=item.obj)


sybil = Sybil(
    parsers=[
        DocTestParser(optionflags=ELLIPSIS),
        PythonCodeBlockParser(),
    ],
    patterns=["*.rst", "*.py"],
)

pytest_collect_file = sybil.pytest()



"""Package for ``requests_mock_flask``."""

import re
from http.cookies import SimpleCookie
from types import ModuleType
from typing import TYPE_CHECKING, Any
from urllib.parse import urljoin

import httpretty  # pyright: ignore[reportMissingTypeStubs]
import httpx
import requests_mock
import responses
import respx
import werkzeug

if TYPE_CHECKING:
    import flask
    import requests


_MockObjType = (
    responses.RequestsMock
    | requests_mock.Mocker
    | requests_mock.Adapter
    | respx.MockRouter
    | respx.Router
    | ModuleType
)


def add_flask_app_to_mock(
    mock_obj: _MockObjType,
    flask_app: "flask.Flask",
    base_url: str,
) -> None:
    """
    Make it so that requests sent to the ``base_url`` are forwarded to
    the
    ``Flask`` app, when in the context of the ``mock_obj``.
    """
    transport = httpx.WSGITransport(app=flask_app)

    def respx_side_effect(
        request: httpx.Request,
    ) -> httpx.Response:
        """Side effect for respx."""
        return transport.handle_request(request=request)

    def responses_callback(
        request: "requests.PreparedRequest",
    ) -> tuple[int, dict[str, str], bytes]:
        """Callback for responses."""
        return _responses_callback(request=request, flask_app=flask_app)

    def requests_mock_callback(
        request: "requests_mock.Request",
        context: "requests_mock.Context",
    ) -> str:
        """Callback for requests_mock."""
        return _requests_mock_callback(
            request=request,
            context=context,
            flask_app=flask_app,
        )

    def httpretty_callback(
        request: "httpretty.core.HTTPrettyRequest",
        uri: str,
        headers: dict[str, Any],
    ) -> tuple[int, dict[str, str | int | bool | None], bytes]:
        """Callback for HTTPretty."""
        return _httpretty_callback(
            request=request,
            uri=uri,
            headers=headers,
            flask_app=flask_app,
        )

    for rule in flask_app.url_map.iter_rules():
        # We replace everything inside angle brackets with a regex pattern.
        # For path variables (<path:...>), we use .+ to match any characters
        # including slashes. For all other variable types, we use [^/]+ to
        # match only within a single path segment.
        path_to_match = re.sub(
            pattern="<path:.+?>",
            repl=".+",
            string=rule.rule,
        )
        path_to_match = re.sub(
            pattern="<.+?>",
            repl="[^/]+",
            string=path_to_match,
        )
        pattern = urljoin(base=base_url, url=path_to_match)
        urls = (re.compile(pattern=pattern), re.compile(pattern=pattern + "$"))

        methods = rule.methods or set()
        for method in methods:
            for url in urls:
                if isinstance(mock_obj, responses.RequestsMock) or (
                    isinstance(mock_obj, ModuleType)
                    and mock_obj.__name__ == "responses"
                ):
                    mock_obj.add_callback(
                        method=method,
                        url=url,
                        callback=responses_callback,
                        content_type=None,
                    )
                elif isinstance(
                    mock_obj, (requests_mock.Mocker | requests_mock.Adapter)
                ):
                    mock_obj.register_uri(
                        method=method,
                        url=url,
                        text=requests_mock_callback,
                    )
                elif isinstance(mock_obj, (respx.MockRouter, respx.Router)):
                    mock_obj.route(
                        method=method,
                        url__regex=url.pattern,
                    ).mock(side_effect=respx_side_effect)
                elif mock_obj.__name__ == "httpretty":
                    httpretty.register_uri(  # type: ignore[no-untyped-call]  # pyright: ignore[reportUnknownMemberType]
                        method=method,
                        uri=url,
                        body=httpretty_callback,  # pyright: ignore[reportArgumentType]
                        forcing_headers={"Content-Type": None},
                    )
                else:
                    msg = (
                        "Expected a HTTPretty, ``requests_mock``, "
                        "``respx``, or ``responses`` object, got "
                        f"module '{mock_obj.__name__}'."
                    )
                    raise TypeError(msg)


def _responses_callback(
    request: "requests.PreparedRequest",
    flask_app: "flask.Flask",
) -> tuple[int, dict[str, str], bytes]:
    """Given a request to the flask app, send an equivalent request to an
    in
    memory fake of the flask app and return some key details of the
    response.

    :param request: The incoming request to pass onto the flask app.
    :param flask_app: The Flask application to pass requests to.
    :return: A tuple of status code, response headers and response data
        from the flask app.
    """
    test_client = flask_app.test_client()
    # See parameters at
    # https://werkzeug.palletsprojects.com/en/0.15.x/test/#werkzeug.test.EnvironBuilder
    cookie_dict = dict(request.headers)
    cookie_string = cookie_dict.get("Cookie", "")
    cookie_list = cookie_string.split(sep=";")
    cookie_list_no_empty = [item for item in cookie_list if item]
    simple_cookie: SimpleCookie = SimpleCookie()
    for cookie in cookie_list_no_empty:
        simple_cookie.load(rawdata=cookie)

    cookies_dict = {k: v.value for k, v in simple_cookie.items()}

    for key, value in cookies_dict.items():
        test_client.set_cookie(
            domain="localhost",
            key=key,
            value=value,
        )

    environ_overrides: dict[str, str] = {}
    if "Content-Length" in request.headers:
        environ_overrides["CONTENT_LENGTH"] = request.headers["Content-Length"]

    headers_dict = dict(request.headers).items()
    environ_builder = werkzeug.test.EnvironBuilder(
        path=request.path_url,
        method=str(object=request.method),
        data=request.body,
        headers=headers_dict,
        environ_overrides=environ_overrides,
    )

    response = test_client.open(environ_builder.get_request())

    return (response.status_code, dict(response.headers), bytes(response.data))


def _httpretty_callback(
    request: "httpretty.core.HTTPrettyRequest",
    uri: str,
    headers: dict[str, Any],
    flask_app: "flask.Flask",
) -> tuple[int, dict[str, str | int | bool | None], bytes]:
    """Given a request to the Flask app, send an equivalent request to an
    in
    memory fake of the Flask app and return some key details of the
    response.

    :param request: The incoming request to pass onto the flask app.
    :param uri: The URI of the request.
    :param headers: The headers of the request.
    :param flask_app: The Flask application to pass requests to.
    :return: A tuple of status code, response headers and response data
        from the flask app.
    """
    # We do this to satisfy linters.
    # The parameters are given to httpretty callbacks, but we do not use them.
    del uri
    del headers

    test_client = flask_app.test_client()
    # See parameters at
    # https://werkzeug.palletsprojects.com/en/0.15.x/test/#werkzeug.test.EnvironBuilder
    cookie_string = request.headers.get(name="Cookie", failobj="")
    cookie_list = cookie_string.split(sep=";")
    cookie_list_no_empty = [item for item in cookie_list if item]
    simple_cookie: SimpleCookie = SimpleCookie()
    for cookie in cookie_list_no_empty:
        simple_cookie.load(rawdata=cookie)

    cookies_dict = {k: v.value for k, v in simple_cookie.items()}

    for key, value in cookies_dict.items():
        test_client.set_cookie(
            domain=request.host,
            key=key,
            value=value,
        )

    environ_overrides: dict[str, str] = {}
    if "Content-Length" in request.headers:
        environ_overrides["CONTENT_LENGTH"] = request.headers["Content-Length"]

    environ_builder = werkzeug.test.EnvironBuilder(
        path=str(object=request.path),
        method=request.method,
        headers=request.headers.items(),
        data=request.body,
        environ_overrides=environ_overrides,
    )
    response = test_client.open(environ_builder.get_request())

    return (response.status_code, dict(response.headers), response.data)


def _requests_mock_callback(
    request: "requests_mock.Request",
    context: "requests_mock.Context",
    flask_app: "flask.Flask",
) -> str:
    """Given a request to the Flask app, send an equivalent request to an
    in
    memory fake of the Flask app and return some key details of the
    response.

    :param request: The incoming request to pass onto the flask app.
    :param context: An object containing the collected known data about
        this response.
    :param flask_app: The Flask application to pass requests to.
    :return: A tuple of status code, response headers and response data
        from the flask app.
    """
    test_client = flask_app.test_client()
    # See parameters at
    # https://werkzeug.palletsprojects.com/en/0.15.x/test/#werkzeug.test.EnvironBuilder
    cookie_string = request.headers.get("Cookie", "")
    cookie_list = cookie_string.split(";")
    cookie_list_no_empty = [item for item in cookie_list if item]
    simple_cookie: SimpleCookie = SimpleCookie()
    for cookie in cookie_list_no_empty:
        simple_cookie.load(rawdata=cookie)

    cookies_dict = {k: v.value for k, v in simple_cookie.items()}

    for key, value in cookies_dict.items():
        test_client.set_cookie(
            domain="localhost",
            key=key,
            value=value,
        )

    environ_overrides: dict[str, str] = {}
    if "Content-Length" in request.headers:
        environ_overrides["CONTENT_LENGTH"] = request.headers["Content-Length"]
    environ_builder = werkzeug.test.EnvironBuilder(
        path=request.path_url,
        method=request.method,
        headers=list(request.headers.items()),
        data=request.body,
        environ_overrides=environ_overrides,
    )
    response = test_client.open(environ_builder.get_request())

    context.headers = dict(response.headers)
    context.status_code = response.status_code
    return str(object=response.data.decode())



"""Tests for the ``requests_mock_flask`` package.

Test with a bunch of route types as per:
https://flask.palletsprojects.com/en/1.1.x/quickstart/#variable-rules
"""

import json
import uuid
from collections.abc import Callable
from contextlib import AbstractContextManager
from functools import partial
from http import HTTPStatus
from types import ModuleType
from typing import Final

import httpretty  # pyright: ignore[reportMissingTypeStubs]
import httpx
import pytest
import requests
import requests_mock
import responses
import respx
import werkzeug
from flask import Flask, Response, jsonify, make_response, request
from requests_mock.exceptions import NoMockAddress
from respx.models import AllMockedAssertionError

from requests_mock_flask import add_flask_app_to_mock

# We use a high timeout to allow interactive debugging while requests are being
# made.
_TIMEOUT_SECONDS: Final[int] = 120

_MockObjType = (
    responses.RequestsMock
    | requests_mock.Mocker
    | ModuleType
    | respx.MockRouter
)
_MockCtxManagerYieldType = _MockObjType | None
_MockCtxType = Callable[[], AbstractContextManager[_MockCtxManagerYieldType]]

_MOCK_CTXS: list[_MockCtxType] = [
    partial(responses.RequestsMock, assert_all_requests_are_fired=False),
    requests_mock.Mocker,
    httpretty.httprettized,
    partial(respx.mock, assert_all_called=False),
]

_MOCK_IDS = ["responses", "requests_mock", "httpretty", "respx"]

_MOCK_CTX_MARKER = pytest.mark.parametrize(
    argnames="mock_ctx",
    argvalues=_MOCK_CTXS,
    ids=_MOCK_IDS,
)


def _get_mock_obj(mock_obj: _MockCtxManagerYieldType) -> _MockObjType:
    """Get the mock object, handling the None yield from httpretty."""
    return mock_obj or httpretty


def _do_get(
    *,
    mock_obj: _MockObjType,
    url: str,
    headers: dict[str, str] | None = None,
) -> requests.Response | httpx.Response:
    """Make a GET request via the appropriate HTTP client."""
    if isinstance(mock_obj, (respx.MockRouter, respx.Router)):
        return httpx.get(url=url, headers=headers, timeout=_TIMEOUT_SECONDS)
    return requests.get(
        url=url,
        headers=headers,
        timeout=_TIMEOUT_SECONDS,
    )


def _do_post(
    *,
    mock_obj: _MockObjType,
    url: str,
    cookies: dict[str, str] | None = None,
) -> requests.Response | httpx.Response:
    """Make a POST request via the appropriate HTTP client."""
    if isinstance(mock_obj, (respx.MockRouter, respx.Router)):
        return httpx.post(url=url, cookies=cookies, timeout=_TIMEOUT_SECONDS)
    return requests.post(
        url=url,
        cookies=cookies,
        timeout=_TIMEOUT_SECONDS,
    )


@_MOCK_CTX_MARKER
def test_simple_route(mock_ctx: _MockCtxType) -> None:
    """A simple GET route works."""
    app = Flask(import_name=__name__, static_folder=None)

    @app.route(rule="/")
    def _() -> str:
        """Return a simple message."""
        return "Hello, World!"

    test_client = app.test_client()
    response = test_client.get("/")

    expected_status_code = HTTPStatus.OK
    expected_content_type = "text/html; charset=utf-8"
    expected_data = b"Hello, World!"

    assert response.status_code == expected_status_code
    assert response.headers["Content-Type"] == expected_content_type
    assert response.data == expected_data

    with mock_ctx() as mock_obj:
        mock_obj_to_add = _get_mock_obj(mock_obj=mock_obj)

        add_flask_app_to_mock(
            mock_obj=mock_obj_to_add,
            flask_app=app,
            base_url="http://www.example.com",
        )

        mock_response = _do_get(
            mock_obj=mock_obj_to_add,
            url="http://www.example.com",
        )

    assert mock_response.status_code == expected_status_code
    assert mock_response.headers["Content-Type"] == expected_content_type
    assert mock_response.text == expected_data.decode()


@_MOCK_CTX_MARKER
def test_headers(mock_ctx: _MockCtxType) -> None:
    """Request headers are sent."""
    app = Flask(import_name=__name__, static_folder=None)

    @app.route(rule="/")
    def _() -> str:
        """
        Check that the headers includes {"hello": "world"} and no
        Content-Type.
        """
        assert "Content-Type" not in request.headers
        assert request.headers["hello"] == "world"
        return "Hello, World!"

    test_client = app.test_client()
    response = test_client.get("/", headers={"hello": "world"})

    expected_status_code = HTTPStatus.OK
    expected_content_type = "text/html; charset=utf-8"
    expected_data = b"Hello, World!"

    assert response.status_code == expected_status_code
    assert response.headers["Content-Type"] == expected_content_type
    assert response.data == expected_data

    with mock_ctx() as mock_obj:
        mock_obj_to_add = _get_mock_obj(mock_obj=mock_obj)

        add_flask_app_to_mock(
            mock_obj=mock_obj_to_add,
            flask_app=app,
            base_url="http://www.example.com",
        )

        mock_response = _do_get(
            mock_obj=mock_obj_to_add,
            url="http://www.example.com",
            headers={"hello": "world"},
        )

    assert mock_response.status_code == expected_status_code
    assert mock_response.headers["Content-Type"] == expected_content_type
    assert mock_response.text == expected_data.decode()


@_MOCK_CTX_MARKER
def test_route_with_json(mock_ctx: _MockCtxType) -> None:
    """A route that returns JSON data works."""
    app = Flask(import_name=__name__, static_folder=None)

    @app.route(rule="/")
    def _() -> tuple[Response, int]:
        """Return a simple JSON message."""
        return jsonify({"hello": "world"}), HTTPStatus.CREATED

    test_client = app.test_client()
    response = test_client.get("/")

    expected_status_code = HTTPStatus.CREATED
    expected_content_type = "application/json"
    expected_json = {"hello": "world"}

    assert response.status_code == expected_status_code
    assert response.headers["Content-Type"] == expected_content_type
    assert response.json == expected_json

    with mock_ctx() as mock_obj:
        mock_obj_to_add = _get_mock_obj(mock_obj=mock_obj)

        add_flask_app_to_mock(
            mock_obj=mock_obj_to_add,
            flask_app=app,
            base_url="http://www.example.com",
        )

        mock_response = _do_get(
            mock_obj=mock_obj_to_add,
            url="http://www.example.com",
        )

    assert mock_response.status_code == expected_status_code
    assert mock_response.headers["Content-Type"] == expected_content_type
    assert mock_response.json() == expected_json


@_MOCK_CTX_MARKER
def test_route_with_variable_no_type_given(mock_ctx: _MockCtxType) -> None:
    """A route with a variable works."""
    app = Flask(import_name=__name__, static_folder=None)

    @app.route(rule="/<my_variable>")
    def _(my_variable: str) -> str:
        """Return a simple message which includes the route variable."""
        return "Hello: " + my_variable

    test_client = app.test_client()
    response = test_client.get("/Frasier")

    expected_status_code = HTTPStatus.OK
    expected_content_type = "text/html; charset=utf-8"
    expected_data = b"Hello: Frasier"

    assert response.status_code == expected_status_code
    assert response.headers["Content-Type"] == expected_content_type
    assert response.data == expected_data

    with mock_ctx() as mock_obj:
        mock_obj_to_add = _get_mock_obj(mock_obj=mock_obj)

        add_flask_app_to_mock(
            mock_obj=mock_obj_to_add,
            flask_app=app,
            base_url="http://www.example.com",
        )

        mock_response = _do_get(
            mock_obj=mock_obj_to_add,
            url="http://www.example.com/Frasier",
        )

    assert mock_response.status_code == expected_status_code
    assert mock_response.headers["Content-Type"] == expected_content_type
    assert mock_response.text == expected_data.decode()


@_MOCK_CTX_MARKER
def test_route_with_string_variable(mock_ctx: _MockCtxType) -> None:
    """A route with a string variable works."""
    app = Flask(import_name=__name__, static_folder=None)

    @app.route(rule="/<string:my_variable>")
    def _(my_variable: str) -> str:
        """Return a simple message which includes the route variable."""
        return "Hello: " + my_variable

    test_client = app.test_client()
    response = test_client.get("/Frasier")

    expected_status_code = HTTPStatus.OK
    expected_content_type = "text/html; charset=utf-8"
    expected_data = b"Hello: Frasier"

    assert response.status_code == expected_status_code
    assert response.headers["Content-Type"] == expected_content_type
    assert response.data == expected_data

    with mock_ctx() as mock_obj:
        mock_obj_to_add = _get_mock_obj(mock_obj=mock_obj)

        add_flask_app_to_mock(
            mock_obj=mock_obj_to_add,
            flask_app=app,
            base_url="http://www.example.com",
        )

        mock_response = _do_get(
            mock_obj=mock_obj_to_add,
            url="http://www.example.com/Frasier",
        )

    assert mock_response.status_code == expected_status_code
    assert mock_response.headers["Content-Type"] == expected_content_type
    assert mock_response.text == expected_data.decode()


@_MOCK_CTX_MARKER
def test_route_with_int_variable(mock_ctx: _MockCtxType) -> None:
    """A route with an int variable works."""
    app = Flask(import_name=__name__, static_folder=None)

    @app.route(rule="/<int:my_variable>")
    def _(my_variable: int) -> str:
        """Return a simple message which includes the route variable."""
        return f"Hello: {my_variable + 5}"

    test_client = app.test_client()
    response = test_client.get("/4")

    expected_status_code = HTTPStatus.OK
    expected_content_type = "text/html; charset=utf-8"
    expected_data = b"Hello: 9"

    assert response.status_code == expected_status_code
    assert response.headers["Content-Type"] == expected_content_type
    assert response.data == expected_data

    with mock_ctx() as mock_obj:
        mock_obj_to_add = _get_mock_obj(mock_obj=mock_obj)

        add_flask_app_to_mock(
            mock_obj=mock_obj_to_add,
            flask_app=app,
            base_url="http://www.example.com",
        )

        mock_response = _do_get(
            mock_obj=mock_obj_to_add,
            url="http://www.example.com/4",
        )

    assert mock_response.status_code == expected_status_code
    assert mock_response.headers["Content-Type"] == expected_content_type
    assert mock_response.text == expected_data.decode()


@_MOCK_CTX_MARKER
def test_route_with_float_variable(mock_ctx: _MockCtxType) -> None:
    """A route with a float variable works."""
    app = Flask(import_name=__name__, static_folder=None)

    @app.route(rule="/<float:my_variable>")
    def _(my_variable: float) -> str:
        """Return a simple message which includes the route variable."""
        return f"Hello: {my_variable + 5}"

    test_client = app.test_client()
    response = test_client.get("/4.0")

    expected_status_code = HTTPStatus.OK
    expected_content_type = "text/html; charset=utf-8"
    expected_data = b"Hello: 9.0"

    assert response.status_code == expected_status_code
    assert response.headers["Content-Type"] == expected_content_type
    assert response.data == expected_data

    with mock_ctx() as mock_obj:
        mock_obj_to_add = _get_mock_obj(mock_obj=mock_obj)

        add_flask_app_to_mock(
            mock_obj=mock_obj_to_add,
            flask_app=app,
            base_url="http://www.example.com",
        )

        mock_response = _do_get(
            mock_obj=mock_obj_to_add,
            url="http://www.example.com/4.0",
        )

    assert mock_response.status_code == expected_status_code
    assert mock_response.headers["Content-Type"] == expected_content_type
    assert mock_response.text == expected_data.decode()


@_MOCK_CTX_MARKER
def test_route_with_path_variable_with_slash(mock_ctx: _MockCtxType) -> None:
    """A route with a path variable works."""
    app = Flask(import_name=__name__, static_folder=None)

    @app.route(rule="/<path:my_variable>")
    def _(my_variable: str) -> str:
        """Return a simple message which includes the route variable."""
        return "Hello: " + my_variable

    test_client = app.test_client()
    response = test_client.get("/foo/bar")

    expected_status_code = HTTPStatus.OK
    expected_content_type = "text/html; charset=utf-8"
    expected_data = b"Hello: foo/bar"

    assert response.status_code == expected_status_code
    assert response.headers["Content-Type"] == expected_content_type
    assert response.data == expected_data

    with mock_ctx() as mock_obj:
        mock_obj_to_add = _get_mock_obj(mock_obj=mock_obj)

        add_flask_app_to_mock(
            mock_obj=mock_obj_to_add,
            flask_app=app,
            base_url="http://www.example.com",
        )

        mock_response = _do_get(
            mock_obj=mock_obj_to_add,
            url="http://www.example.com/foo/bar",
        )

    assert mock_response.status_code == expected_status_code
    assert mock_response.headers["Content-Type"] == expected_content_type
    assert mock_response.text == expected_data.decode()


@_MOCK_CTX_MARKER
def test_route_with_string_variable_with_slash(mock_ctx: _MockCtxType) -> None:
    """A route with a string variable when given a slash works."""
    app = Flask(import_name=__name__, static_folder=None)

    @app.route(rule="/<string:my_variable>")
    def _(_: str) -> str:
        """Return an empty string."""
        return ""  # pragma: no cover

    test_client = app.test_client()
    response = test_client.get("/foo/bar")

    expected_status_code = HTTPStatus.NOT_FOUND
    expected_content_type = "text/html; charset=utf-8"

    assert response.status_code == expected_status_code
    assert response.headers["Content-Type"] == expected_content_type
    assert b"not found on the server" in response.data

    with mock_ctx() as mock_obj:
        mock_obj_to_add = _get_mock_obj(mock_obj=mock_obj)

        add_flask_app_to_mock(
            mock_obj=mock_obj_to_add,
            flask_app=app,
            base_url="http://www.example.com",
        )

        mock_response = _do_get(
            mock_obj=mock_obj_to_add,
            url="http://www.example.com/foo/bar",
        )

    assert mock_response.status_code == expected_status_code
    assert mock_response.headers["Content-Type"] == expected_content_type
    assert "not found on the server" in mock_response.text


@_MOCK_CTX_MARKER
def test_route_with_uuid_variable(mock_ctx: _MockCtxType) -> None:
    """A route with a uuid variable works."""
    app = Flask(import_name=__name__, static_folder=None)

    @app.route(rule="/<uuid:my_variable>")
    def _(my_variable: uuid.UUID) -> str:
        """Return a simple message which includes the route variables."""
        return "Hello: " + my_variable.hex

    test_client = app.test_client()
    random_uuid = uuid.uuid4()
    response = test_client.get(f"/{random_uuid}")

    expected_status_code = HTTPStatus.OK
    expected_content_type = "text/html; charset=utf-8"
    expected_data = f"Hello: {random_uuid.hex}".encode()

    assert response.status_code == expected_status_code
    assert response.headers["Content-Type"] == expected_content_type
    assert response.data == expected_data

    with mock_ctx() as mock_obj:
        mock_obj_to_add = _get_mock_obj(mock_obj=mock_obj)

        add_flask_app_to_mock(
            mock_obj=mock_obj_to_add,
            flask_app=app,
            base_url="http://www.example.com",
        )

        mock_response = _do_get(
            mock_obj=mock_obj_to_add,
            url=f"http://www.example.com/{random_uuid}",
        )

    assert mock_response.status_code == expected_status_code
    assert mock_response.headers["Content-Type"] == expected_content_type
    assert mock_response.text == expected_data.decode()


@_MOCK_CTX_MARKER
def test_nested_path(mock_ctx: _MockCtxType) -> None:
    """A route with a variable nested in a path works."""
    app = Flask(import_name=__name__, static_folder=None)

    @app.route(rule="/users/<int:my_variable>/posts")
    def _(my_variable: int) -> str:
        """Return a simple message which includes the route variable."""
        return f"Posts for: {my_variable}"

    test_client = app.test_client()
    response = test_client.get("/users/4/posts")

    expected_status_code = HTTPStatus.OK
    expected_content_type = "text/html; charset=utf-8"
    expected_data = b"Posts for: 4"

    assert response.status_code == expected_status_code
    assert response.headers["Content-Type"] == expected_content_type
    assert response.data == expected_data

    with mock_ctx() as mock_obj:
        mock_obj_to_add = _get_mock_obj(mock_obj=mock_obj)

        add_flask_app_to_mock(
            mock_obj=mock_obj_to_add,
            flask_app=app,
            base_url="http://www.example.com",
        )

        mock_response = _do_get(
            mock_obj=mock_obj_to_add,
            url="http://www.example.com/users/4/posts",
        )

    assert mock_response.status_code == expected_status_code
    assert mock_response.headers["Content-Type"] == expected_content_type
    assert mock_response.text == expected_data.decode()


@_MOCK_CTX_MARKER
def test_route_with_multiple_variables(mock_ctx: _MockCtxType) -> None:
    """A route with multiple variables works."""
    app = Flask(import_name=__name__, static_folder=None)

    @app.route(rule="/users/<string:my_org>/<string:my_user>/posts")
    def _(my_org: str, my_user: str) -> str:
        """Return a simple message which includes the route variables."""
        return "Posts for: " + my_org + "/" + my_user

    test_client = app.test_client()
    response = test_client.get("/users/cranes/frasier/posts")

    expected_status_code = HTTPStatus.OK
    expected_content_type = "text/html; charset=utf-8"
    expected_data = b"Posts for: cranes/frasier"

    assert response.status_code == expected_status_code
    assert response.headers["Content-Type"] == expected_content_type
    assert response.data == expected_data

    with mock_ctx() as mock_obj:
        mock_obj_to_add = _get_mock_obj(mock_obj=mock_obj)

        add_flask_app_to_mock(
            mock_obj=mock_obj_to_add,
            flask_app=app,
            base_url="http://www.example.com",
        )

        mock_response = _do_get(
            mock_obj=mock_obj_to_add,
            url="http://www.example.com/users/cranes/frasier/posts",
        )

    assert mock_response.status_code == expected_status_code
    assert mock_response.headers["Content-Type"] == expected_content_type
    assert mock_response.text == expected_data.decode()


@_MOCK_CTX_MARKER
def test_post_verb(mock_ctx: _MockCtxType) -> None:
    """A route with the POST verb works."""
    app = Flask(import_name=__name__, static_folder=None)

    @app.route(rule="/", methods=["POST"])
    def _() -> str:
        """Return a simple message."""
        return "Hello, World!"

    test_client = app.test_client()
    response = test_client.post("/")

    expected_status_code = HTTPStatus.OK
    expected_content_type = "text/html; charset=utf-8"
    expected_data = b"Hello, World!"

    assert response.status_code == expected_status_code
    assert response.headers["Content-Type"] == expected_content_type
    assert response.data == expected_data

    with mock_ctx() as mock_obj:
        mock_obj_to_add = _get_mock_obj(mock_obj=mock_obj)

        add_flask_app_to_mock(
            mock_obj=mock_obj_to_add,
            flask_app=app,
            base_url="http://www.example.com",
        )

        mock_response = _do_post(
            mock_obj=mock_obj_to_add,
            url="http://www.example.com/",
        )

    assert mock_response.status_code == expected_status_code
    assert mock_response.headers["Content-Type"] == expected_content_type
    assert mock_response.text == expected_data.decode()


@pytest.mark.parametrize(
    argnames="custom_content_length",
    argvalues=["1", "100"],
)
@_MOCK_CTX_MARKER
def test_incorrect_content_length(
    *,
    custom_content_length: str,
    mock_ctx: _MockCtxType,
) -> None:
    """
    Custom content length headers are passed through to the Flask
    endpoint.
    """
    app = Flask(import_name=__name__, static_folder=None)
    data = b"12345"

    @app.route(rule="/", methods=["POST"])
    def _() -> str:
        """Check some features of the request."""
        request.environ["wsgi.input_terminated"] = True
        assert len(data) == len(request.data)
        assert request.headers["Content-Length"] == custom_content_length
        return ""

    test_client = app.test_client()
    environ_builder = werkzeug.test.EnvironBuilder(
        path="/",
        method="POST",
        data=b"12345",
        environ_overrides={"CONTENT_LENGTH": custom_content_length},
    )

    response = test_client.open(environ_builder.get_request())

    expected_status_code = HTTPStatus.OK

    assert response.status_code == expected_status_code

    with mock_ctx() as mock_obj:
        mock_obj_to_add = _get_mock_obj(mock_obj=mock_obj)

        add_flask_app_to_mock(
            mock_obj=mock_obj_to_add,
            flask_app=app,
            base_url="http://www.example.com",
        )

        mock_response: requests.Response | httpx.Response
        if isinstance(mock_obj_to_add, (respx.MockRouter, respx.Router)):
            mock_response = httpx.request(
                method="POST",
                url="http://www.example.com/",
                content=data,
                headers={"Content-Length": custom_content_length},
            )
        else:
            requests_request = requests.Request(
                method="POST",
                url="http://www.example.com/",
                data=data,
            ).prepare()
            requests_request.headers["Content-Length"] = custom_content_length
            session = requests.Session()
            mock_response = session.send(request=requests_request)

    assert mock_response.status_code == expected_status_code


@_MOCK_CTX_MARKER
def test_multiple_http_verbs(mock_ctx: _MockCtxType) -> None:
    """A route with multiple verbs works."""
    app = Flask(import_name=__name__, static_folder=None)

    @app.route(rule="/", methods=["GET", "POST"])
    def _() -> str:
        """Return a simple message."""
        return "Hello, World!"

    test_client = app.test_client()
    get_response = test_client.get("/")
    post_response = test_client.post("/")

    expected_status_code = HTTPStatus.OK
    expected_content_type = "text/html; charset=utf-8"
    expected_data = b"Hello, World!"

    assert get_response.status_code == expected_status_code
    assert get_response.headers["Content-Type"] == expected_content_type
    assert get_response.data == expected_data

    assert post_response.status_code == expected_status_code
    assert post_response.headers["Content-Type"] == expected_content_type
    assert post_response.data == expected_data

    with mock_ctx() as mock_obj:
        mock_obj_to_add = _get_mock_obj(mock_obj=mock_obj)

        add_flask_app_to_mock(
            mock_obj=mock_obj_to_add,
            flask_app=app,
            base_url="http://www.example.com",
        )

        mock_get_response = _do_get(
            mock_obj=mock_obj_to_add,
            url="http://www.example.com/",
        )
        mock_post_response = _do_post(
            mock_obj=mock_obj_to_add,
            url="http://www.example.com/",
        )

    assert mock_get_response.status_code == expected_status_code
    assert mock_get_response.headers["Content-Type"] == expected_content_type
    assert mock_get_response.text == expected_data.decode()

    assert mock_post_response.status_code == expected_status_code
    assert mock_post_response.headers["Content-Type"] == expected_content_type
    assert mock_post_response.text == expected_data.decode()


@_MOCK_CTX_MARKER
def test_wrong_type_given(mock_ctx: _MockCtxType) -> None:
    """A route with the wrong type given works."""
    app = Flask(import_name=__name__, static_folder=None)

    @app.route(rule="/<int:my_variable>")
    def _(_: int) -> str:
        """Return an empty string."""
        return ""  # pragma: no cover

    test_client = app.test_client()
    response = test_client.get("/a")

    expected_status_code = HTTPStatus.NOT_FOUND
    expected_content_type = "text/html; charset=utf-8"

    assert response.status_code == expected_status_code
    assert response.headers["Content-Type"] == expected_content_type
    assert b"not found on the server" in response.data

    with mock_ctx() as mock_obj:
        mock_obj_to_add = _get_mock_obj(mock_obj=mock_obj)

        add_flask_app_to_mock(
            mock_obj=mock_obj_to_add,
            flask_app=app,
            base_url="http://www.example.com",
        )

        mock_response = _do_get(
            mock_obj=mock_obj_to_add,
            url="http://www.example.com/a",
        )

    assert mock_response.status_code == expected_status_code
    assert mock_response.headers["Content-Type"] == expected_content_type
    assert "not found on the server" in mock_response.text


@_MOCK_CTX_MARKER
def test_405_no_such_method(mock_ctx: _MockCtxType) -> None:
    """A route with the wrong method given works."""
    app = Flask(import_name=__name__, static_folder=None)

    @app.route(rule="/")
    def _() -> str:
        """Return an empty string."""
        return ""  # pragma: no cover

    test_client = app.test_client()
    response = test_client.post("/")

    expected_status_code = HTTPStatus.METHOD_NOT_ALLOWED
    expected_content_type = "text/html; charset=utf-8"

    assert response.status_code == expected_status_code
    assert response.headers["Content-Type"] == expected_content_type
    assert b"not allowed for the requested URL." in response.data

    with mock_ctx() as mock_obj:
        mock_obj_to_add = _get_mock_obj(mock_obj=mock_obj)

        add_flask_app_to_mock(
            mock_obj=mock_obj_to_add,
            flask_app=app,
            base_url="http://www.example.com",
        )

        expected_exceptions: tuple[type[Exception], ...] = (
            AllMockedAssertionError,
            requests.exceptions.ConnectionError,
            NoMockAddress,
            ValueError,
        )
        with pytest.raises(expected_exception=expected_exceptions):
            _do_post(
                mock_obj=mock_obj_to_add,
                url="http://www.example.com/",
            )


@_MOCK_CTX_MARKER
def test_request_needs_content_type(mock_ctx: _MockCtxType) -> None:
    """Routes which require a content type are supported."""
    app = Flask(import_name=__name__, static_folder=None)

    @app.route(rule="/")
    def _() -> str:
        """Check the MIME type and return a simple message."""
        assert request.mimetype == "application/json"
        return "Hello, World!"

    test_client = app.test_client()
    response = test_client.get("/", content_type="application/json")

    expected_status_code = HTTPStatus.OK
    expected_content_type = "text/html; charset=utf-8"
    expected_data = b"Hello, World!"

    assert response.status_code == expected_status_code
    assert response.headers["Content-Type"] == expected_content_type
    assert response.data == expected_data

    with mock_ctx() as mock_obj:
        mock_obj_to_add = _get_mock_obj(mock_obj=mock_obj)

        add_flask_app_to_mock(
            mock_obj=mock_obj_to_add,
            flask_app=app,
            base_url="http://www.example.com",
        )

        mock_response = _do_get(
            mock_obj=mock_obj_to_add,
            url="http://www.example.com",
            headers={"Content-Type": "application/json"},
        )

    assert mock_response.status_code == expected_status_code
    assert mock_response.headers["Content-Type"] == expected_content_type
    assert mock_response.text == expected_data.decode()


@_MOCK_CTX_MARKER
def test_request_needs_data(mock_ctx: _MockCtxType) -> None:
    """Routes which require data are supported."""
    app = Flask(import_name=__name__, static_folder=None)

    @app.route(rule="/")
    def _() -> str:
        """Check the MIME type and return some given data."""
        assert request.mimetype == "application/json"
        request_json = request.get_json()
        return str(object=request_json["hello"])

    test_client = app.test_client()
    response = test_client.get(
        "/",
        content_type="application/json",
        data=json.dumps(obj={"hello": "world"}),
    )

    expected_status_code = HTTPStatus.OK
    expected_content_type = "text/html; charset=utf-8"
    expected_data = b"world"

    assert response.status_code == expected_status_code
    assert response.headers["Content-Type"] == expected_content_type
    assert response.data == expected_data

    with mock_ctx() as mock_obj:
        mock_obj_to_add = _get_mock_obj(mock_obj=mock_obj)

        add_flask_app_to_mock(
            mock_obj=mock_obj_to_add,
            flask_app=app,
            base_url="http://www.example.com",
        )

        mock_response: requests.Response | httpx.Response
        if isinstance(mock_obj_to_add, (respx.MockRouter, respx.Router)):
            mock_response = httpx.request(
                method="GET",
                url="http://www.example.com",
                headers={"Content-Type": "application/json"},
                content=json.dumps(obj={"hello": "world"}).encode(),
            )
        else:
            mock_response = requests.get(
                url="http://www.example.com",
                headers={"Content-Type": "application/json"},
                data=json.dumps(obj={"hello": "world"}),
                timeout=_TIMEOUT_SECONDS,
            )

    assert mock_response.status_code == expected_status_code
    assert mock_response.headers["Content-Type"] == expected_content_type
    assert mock_response.text == expected_data.decode()


@_MOCK_CTX_MARKER
def test_multiple_functions_same_path_different_type(
    mock_ctx: _MockCtxType,
) -> None:
    """
    When multiple functions exist with the same path but have a
    different type,
    the mock matches them just the same.
    """
    app = Flask(import_name=__name__, static_folder=None)

    def show_type(variable: float | str) -> str:
        """Return a string which includes the type of the variable."""
        return f"{variable}, {type(variable)}"

    app.add_url_rule(rule="/<variable>", view_func=show_type)
    app.add_url_rule(rule="/<int:variable>", view_func=show_type)
    app.add_url_rule(rule="/<string:variable>", view_func=show_type)

    test_client = app.test_client()
    response = test_client.get("/4")

    expected_status_code = HTTPStatus.OK
    expected_content_type = "text/html; charset=utf-8"
    expected_data = b"4, <class 'int'>"

    assert response.status_code == expected_status_code
    assert response.headers["Content-Type"] == expected_content_type
    assert response.data == expected_data

    with mock_ctx() as mock_obj:
        mock_obj_to_add = _get_mock_obj(mock_obj=mock_obj)

        add_flask_app_to_mock(
            mock_obj=mock_obj_to_add,
            flask_app=app,
            base_url="http://www.example.com",
        )

        mock_response = _do_get(
            mock_obj=mock_obj_to_add,
            url="http://www.example.com/4",
        )

    assert mock_response.status_code == expected_status_code
    assert mock_response.headers["Content-Type"] == expected_content_type
    assert mock_response.text == expected_data.decode()


@_MOCK_CTX_MARKER
def test_query_string(mock_ctx: _MockCtxType) -> None:
    """Query strings work."""
    app = Flask(import_name=__name__, static_folder=None)

    @app.route(rule="/")
    def _() -> str:
        """
        Return a simple message which includes a request query
        parameter.
        """
        result = request.args["frasier"]
        return f"Hello: {result}"

    test_client = app.test_client()
    response = test_client.get("/?frasier=crane")

    expected_status_code = HTTPStatus.OK
    expected_content_type = "text/html; charset=utf-8"
    expected_data = b"Hello: crane"

    assert response.status_code == expected_status_code
    assert response.headers["Content-Type"] == expected_content_type
    assert response.data == expected_data

    with mock_ctx() as mock_obj:
        mock_obj_to_add = _get_mock_obj(mock_obj=mock_obj)

        add_flask_app_to_mock(
            mock_obj=mock_obj_to_add,
            flask_app=app,
            base_url="http://www.example.com",
        )

        mock_response = _do_get(
            mock_obj=mock_obj_to_add,
            url="http://www.example.com?frasier=crane",
        )

    assert mock_response.status_code == expected_status_code
    assert mock_response.headers["Content-Type"] == expected_content_type
    assert mock_response.text == expected_data.decode()


@_MOCK_CTX_MARKER
def test_cookies(mock_ctx: _MockCtxType) -> None:
    """Cookies work."""
    app = Flask(import_name=__name__, static_folder=None)

    @app.route(rule="/", methods=["POST"])
    def _() -> Response:
        """Set cookies and return a simple message."""
        response = make_response()
        response.set_cookie(key="frasier_set", value="crane_set")
        assert request.cookies, request
        assert request.cookies["frasier"] == "crane"
        assert request.cookies["frasier2"] == "crane2"
        response.data = "Hello, World!"
        return response

    test_client = app.test_client()
    test_client.set_cookie(
        domain="localhost",
        key="frasier",
        value="crane",
    )
    test_client.set_cookie(
        domain="localhost",
        key="frasier2",
        value="crane2",
    )
    response = test_client.post("/")

    expected_status_code = HTTPStatus.OK
    expected_content_type = "text/html; charset=utf-8"
    expected_data = b"Hello, World!"

    assert response.status_code == expected_status_code, response.data
    assert response.headers["Content-Type"] == expected_content_type
    new_cookie = test_client.get_cookie(key="frasier_set")
    assert new_cookie is not None
    assert new_cookie.key == "frasier_set"
    assert new_cookie.value == "crane_set"
    assert response.data == expected_data

    with mock_ctx() as mock_obj:
        mock_obj_to_add = _get_mock_obj(mock_obj=mock_obj)

        add_flask_app_to_mock(
            mock_obj=mock_obj_to_add,
            flask_app=app,
            base_url="http://www.example.com",
        )

        mock_response = _do_post(
            mock_obj=mock_obj_to_add,
            url="http://www.example.com",
            cookies={
                "frasier": "crane",
                "frasier2": "crane2",
            },
        )

    assert mock_response.status_code == expected_status_code
    assert mock_response.headers["Content-Type"] == expected_content_type
    assert mock_response.text == expected_data.decode()
    assert mock_response.cookies["frasier_set"] == "crane_set"


@_MOCK_CTX_MARKER
def test_no_content_type(mock_ctx: _MockCtxType) -> None:
    """It is possible to get a response without a content type."""
    app = Flask(import_name=__name__, static_folder=None)

    @app.route(rule="/", methods=["GET"])
    def _() -> Response:
        """Return a simple message with no Content-Type."""
        response = make_response()
        response.data = "Hello, World!"
        del response.headers["Content-Type"]
        return response

    test_client = app.test_client()
    response = test_client.get("/")

    expected_status_code = HTTPStatus.OK
    expected_data = b"Hello, World!"

    assert response.status_code == expected_status_code
    assert "Content-Type" not in response.headers
    assert response.data == expected_data

    with mock_ctx() as mock_obj:
        mock_obj_to_add = _get_mock_obj(mock_obj=mock_obj)

        add_flask_app_to_mock(
            mock_obj=mock_obj_to_add,
            flask_app=app,
            base_url="http://www.example.com",
        )

        mock_response = _do_get(
            mock_obj=mock_obj_to_add,
            url="http://www.example.com",
        )

    assert mock_response.status_code == expected_status_code
    assert "Content-Type" not in mock_response.headers
    assert mock_response.text == expected_data.decode()


@_MOCK_CTX_MARKER
def test_overlapping_routes_multiple_requests(mock_ctx: _MockCtxType) -> None:
    """
    A route with overlap to another route works across multiple
    requests.
    """
    app = Flask(import_name=__name__, static_folder=None)

    def base_route() -> str:
        """Return a simple message."""
        return "Hello: World"

    def with_variable(my_variable: str) -> str:
        """Return a simple message which includes the route variable."""
        return "Hello: " + my_variable

    app.add_url_rule(rule="/base", methods=["GET"], view_func=base_route)
    app.add_url_rule(
        rule="/base/<string:my_variable>",
        methods=["GET"],
        view_func=with_variable,
    )

    test_client = app.test_client()
    response = test_client.get("/base")

    expected_status_code = HTTPStatus.OK
    expected_content_type = "text/html; charset=utf-8"
    expected_base_data = b"Hello: World"

    assert response.status_code == expected_status_code
    assert response.data == expected_base_data

    test_client = app.test_client()
    response = test_client.get("/base/Frasier")

    expected_var_data = b"Hello: Frasier"

    assert response.status_code == expected_status_code
    assert response.data == expected_var_data

    with mock_ctx() as mock_obj:
        mock_obj_to_add = _get_mock_obj(mock_obj=mock_obj)

        add_flask_app_to_mock(
            mock_obj=mock_obj_to_add,
            flask_app=app,
            base_url="http://www.example.com",
        )

        mock_response_base = _do_get(
            mock_obj=mock_obj_to_add,
            url="http://www.example.com/base",
        )

        mock_response_var = _do_get(
            mock_obj=mock_obj_to_add,
            url="http://www.example.com/base/Frasier",
        )

        mock_response_base_2 = _do_get(
            mock_obj=mock_obj_to_add,
            url="http://www.example.com/base",
        )

    assert mock_response_base.status_code == expected_status_code
    assert mock_response_base.headers["Content-Type"] == expected_content_type
    assert mock_response_base.text == expected_base_data.decode()

    assert mock_response_var.status_code == expected_status_code
    assert mock_response_var.headers["Content-Type"] == expected_content_type
    assert mock_response_var.text == expected_var_data.decode()

    assert mock_response_base_2.status_code == expected_status_code
    assert (
        mock_response_base_2.headers["Content-Type"] == expected_content_type
    )
    assert mock_response_base_2.text == expected_base_data.decode()


@_MOCK_CTX_MARKER
def test_multiple_variables_no_extra_segments(mock_ctx: _MockCtxType) -> None:
    """
    A route with multiple variables should not match URLs with extra
    segments.
    """
    app = Flask(import_name=__name__, static_folder=None)

    @app.route(rule="/users/<string:my_org>/<string:my_user>/posts")
    def _(my_org: str, my_user: str) -> str:
        """Return a simple message which includes the route variables."""
        return "Posts for: " + my_org + "/" + my_user

    # Verify the real Flask app rejects URLs with extra segments
    test_client = app.test_client()
    response = test_client.get("/users/cranes/frasier/extra/posts")
    assert response.status_code == HTTPStatus.NOT_FOUND

    with mock_ctx() as mock_obj:
        mock_obj_to_add = _get_mock_obj(mock_obj=mock_obj)

        add_flask_app_to_mock(
            mock_obj=mock_obj_to_add,
            flask_app=app,
            base_url="http://www.example.com",
        )

        # Verify that the correct URL works
        valid_response = _do_get(
            mock_obj=mock_obj_to_add,
            url="http://www.example.com/users/cranes/frasier/posts",
        )
        assert valid_response.status_code == HTTPStatus.OK
        assert valid_response.text == "Posts for: cranes/frasier"


# This test cannot use httpretty because httpretty has a bug with
# allow_net_connect=False when used with urllib3 2.3.0+.
# See: https://github.com/gabrielfalcao/HTTPretty/issues/484
_MOCK_CTXS_NO_HTTPRETTY: list[_MockCtxType] = [
    ctx
    for ctx, mock_id in zip(_MOCK_CTXS, _MOCK_IDS, strict=True)
    if mock_id != "httpretty"
]

_MOCK_IDS_NO_HTTPRETTY = [
    mock_id for mock_id in _MOCK_IDS if mock_id != "httpretty"
]

_MOCK_CTX_MARKER_NO_HTTPRETTY = pytest.mark.parametrize(
    argnames="mock_ctx",
    argvalues=_MOCK_CTXS_NO_HTTPRETTY,
    ids=_MOCK_IDS_NO_HTTPRETTY,
)


@_MOCK_CTX_MARKER_NO_HTTPRETTY
def test_multiple_variables_rejects_extra_segments(
    mock_ctx: _MockCtxType,
) -> None:
    """
    URLs with extra path segments should not match routes with multiple
    variables.
    """
    app = Flask(import_name=__name__, static_folder=None)

    @app.route(rule="/users/<string:my_org>/<string:my_user>/posts")
    def _(my_org: str, my_user: str) -> str:
        """Return a simple message which includes the route variables."""
        return "Posts for: " + my_org + "/" + my_user  # pragma: no cover

    with mock_ctx() as mock_obj:
        mock_obj_to_add = _get_mock_obj(mock_obj=mock_obj)
        add_flask_app_to_mock(
            mock_obj=mock_obj_to_add,
            flask_app=app,
            base_url="http://www.example.com",
        )

        expected_exceptions: tuple[type[Exception], ...] = (
            AllMockedAssertionError,
            requests.exceptions.ConnectionError,
            NoMockAddress,
        )
        with pytest.raises(expected_exception=expected_exceptions):
            _do_get(
                mock_obj=mock_obj_to_add,
                url="http://www.example.com/users/cranes/frasier/extra/posts",
            )


def test_unknown_mock_module() -> None:
    """When an unknown mock module is passed in, an error is raised."""
    app = Flask(import_name=__name__, static_folder=None)

    @app.route(rule="/")
    def _() -> str:
        """Return a simple message."""
        return ""  # pragma: no cover

    expected_error = (
        "Expected a HTTPretty, ``requests_mock``, "
        "``respx``, or ``responses`` object, "
        "got module 'json'."
    )
    with pytest.raises(expected_exception=TypeError, match=expected_error):
        add_flask_app_to_mock(
            mock_obj=json,
            flask_app=app,
            base_url="http://www.example.com",
        )
